"""
Constants
=========

Those constants are generic and can be used across the different datastore
types.
"""

SORT_ASCENDING = 2
SORT_DESCENDING = 1

ERROR_CODE_VALIDATION = 'validation_errors'
ERROR_CODE_DUPLICATE_KEY = 'duplicate_key'
